# -*- coding: utf-8 -*-
"""Calculations related to the bandstructure.

This module contains the 'BandstructureModel' class, which handles all things
related to the bandstructure.
"""

# Third Party Imports
import numpy as np
import matplotlib.pyplot as plt


class BandstructureModel:
    """Bandstructure model handles all things related to the bandstructure.

    Should not be used directly by the user but called by the abstraction layer
    octopost class.

    Args:
        octopost (object): An instance of the octopost class.
    """

    def __init__(self, ocotpost):
        self.op = ocotpost

    def get_bandstructure(self, plot=False):
        """Returns the bandstructure.

        Reads the 'bandstructure' file found in 'static/' and returns the path
        along the k points as coordinate and (kx, ky, kz) as well as the bands.
        Optionally a pre-made standard plot can be returned as well.

        Args:
            plot (bool): Returns matplotlib handles for a basic plot of the
            bandstructure.

        Returns:
            nd.array:
                Numpy array with all k-points. First column is
                the coordinate, remaining columns are kx, ky, kz.
            nd.array:
                Numpy array with each column being one band.
            fig, ax:
                Matplotlib handles for the plot. Only returned if
                plot==True.
        """

        file = 'bandstructure'
        dir_ = 'static'
        path = self.op._check_for_file(dir_, file)

        with open(path, 'r') as f:
            data = np.loadtxt(f)

        # First column == coordinate along path in k-space
        # Second to Fourth column = (kx, ky, kz) of coordinate in k-space
        # Fith column onwards = individual bands
        k_path = data[:, :4]
        bands = data[:, 4:]

        if plot:
            fig, ax = plt.subplots()
            for i in range(0, bands.shape[1]):
                ax.plot(k_path[:, 0], bands[:, i])

            ax.set_xlabel('$k$ (1/bohr)', fontsize=14)
            ax.set_ylabel('$E$ (Hartree)', fontsize=14)

            return k_path, bands, (fig, ax)

        else:
            return k_path, bands



# -*- coding: utf-8 -*-
"""Definition of various useful code snipets.
"""

# Third Party Imports
import numpy as np
from vtk import vtkStructuredGridReader
from vtk.util.numpy_support import vtk_to_numpy


def cartesian_to_spherical(x, y, z):
    """Converts cartesian coordinates to spherical ones::

        r = sqrt(x^2 + y^2 + z^2)
        theta = arccos(z / r)
        phi = arctan(y / x)

    Args:
        x (nd.array or float): x-Coordinate.
        y (nd.array or float): y-Coordinate.
        z (nd.array or float): z-Coordinate.

    Returns:
        nd.array or float:
            Radius.
        nd.array or float:
            Polar angle theta.
        nd.array or float:
            Azimuthal angle phi.
    """

    r = np.sqrt(x**2 + y**2 + z**2)

    theta = np.arccos(z / r)
    theta[np.isnan(theta)] = 0

    phi = np.arctan2(y, x)
    phi[phi < 0] += 2 * np.pi

    return r, theta, phi


def k_to_E(k):
    """Converts momentum to (kinetic) energy of an electron::

        E = m * v^2 / 2 = m * (k / m)^2 / 2 = k^2 / (2 * m)

    Args:
        k (nd.array or float): Momentum to be converted in atomic units.

    Returns:
        nd.array or float:
            Energy in atomic units.
    """

    return k**2 / 2


def E_to_k(E):
    """Converts (kinetic) energy (of an electron) to momentum::

        k = m * v = m * sqrt(2 * E / m) = sqrt(2 * E * m)

    Args:
        E (nd.array or float): Energy to be converted in atomic units.

    Returns:
        nd.array or float:
            Momentum in atomic units.
    """

    return np.sqrt(2 * E)


def find_nearest_index(array, value):
    """Finds the index of the closes element to 'value'.

    Args:
        array (nd.array): Array of values.
        value (float): Value to be located inside the array.

    Returns:
        int:
            Index of the element inside 'array' that is closest to 'value'.
    """

    idx = (np.abs(array - value)).argmin()
    return idx


def read_vtk_file(path, array_name):
    """Reads a .vtk file containing a structured grid.

    Args:
        path (str): Absolute path to the file that is to be read.
        array_name (str): Name of the array inside the file to be read.

    Returns:
        nd.array:
            4D array. Last dimension holds length 3 tuple with
            x, y and z coordinate for that point.
        nd.array:
            3D array with data at each point.
    """

    reader = vtkStructuredGridReader()
    reader.SetFileName(str(path))
    reader.Update()

    grid = reader.GetOutput().GetPoints().GetData()
    data = reader.GetOutput().GetPointData().GetArray(array_name)
    shape = np.array(reader.GetOutput().GetDimensions())
    grid = vtk_to_numpy(grid).reshape((*shape, 3), order='F')
    data = vtk_to_numpy(data).reshape(shape, order='F')

    return grid, data



# -*- coding: utf-8 -*-
"""Definition of various useful units used for conversion.

Source for all values unless specified otherwise:
https://physics.nist.gov/cuu/Constants/index.html

Attributes:
    bohr (float): Bohr atomic radius in [Angstrom].
    hartree (float): Hartree in [eV].
    m_e (float): Electron mass in [kg].
    e (float): Elementary charge in [Coulomb].
"""

# Third Party Imports
import numpy as np

bohr = 0.529177210903
hartree = 27.211386245988
m_e = 9.1093837015e-31
e = 1.602176634e-19


def energy(data, from_='hartree', to='eV'):
    """Converts energy units.

    Args:
        data (float or list): The initual value to be converted in units
            'from\_'.
        from\_ (str): The unit of the values supplied. Available options are:
            'hartree' (= Hartree), 'eV' (= Electron Volt).
        to (str): The unit of the output. Available options are:
            'hartree', 'eV'.

    Returns:
        float or list:
            The data but converted from 'from\_' to 'to'.
    """

    available_units = ('hartree', 'eV')
    if from_ not in available_units or to not in available_units:
        print('Unit requested or supplied not supported.')
        raise ValueError

    if from_ == 'hartree':
        return np.array(data) * hartree

    else:
        return np.array(data) / hartree


def inverse_length(data, from_='bohr', to='A'):
    """Converts inverse length (reciprocal distance).

    Args:
        data (float or list): The initual value to be converted in units
            'from\_'.
        from\_ (str): The unit of the values supplied. Available options are:
            'bohr' (= 1/Bohr radius), 'A' (= 1/Angstrom).
        to (str): The unit of the output. Available options are:
            'bohr' (= 1/Bohr radius), 'A' (= 1/Angstrom).

    Returns:
        float or list:
            The data but converted from 'from\_' to 'to'.
    """

    available_units = ('bohr', 'A')
    if from_ not in available_units or to not in available_units:
        print('Unit requested or supplied not supported.')
        raise ValueError

    if from_ == 'bohr':
        return np.array(data) / bohr

    else:
        return np.array(data) * bohr



# -*- coding: utf-8 -*-
"""Convienence class to extract various data from Octopus output.
"""

# Python Imports
import re

# Third Party Imports
import numpy as np
import matplotlib.pyplot as plt

# Octopost Imports
import units


class OutputReadModel:
    """OutputRead model handles all things related to basic information output.

    Should not be used directly by the user but called by the abstraction layer
    octopost class.

    Args:
        octopost (object): An instance of the octopost class.
    """

    def __init__(self, ocotpost):
        self.op = ocotpost

    def get_fermi(self, energy_units='hartree'):
        """Returns fermi energy found in the 'total-dos-efermi.dat' file.

        Args:
            energy_units (str): Specify the units in which the energy is
                returned. Options: 'hartree', 'eV'

        Returns:
            float:
                Fermi energy in the specified units.
        """

        file = 'total-dos-efermi.dat'
        dir_ = 'static'
        path = self.op._check_for_file(dir_, file)

        with open(path, 'r') as f:
            # Fermi energy is the first value in the file
            fermi = np.loadtxt(f)[0, 0]

        if energy_units == 'eV':
            fermi = units.energy(fermi, from_='hartree', to='eV')

        return fermi

    def get_eigenvalues(self, n_max=None, n_min=1, k_point=None,
                        energy_units='hartree'):
        """Returns the eigenvalues found in the 'info' file.

        Args:
            n_max (int): How many eigenvalues to be returned. If 'None' all of
                them are returned. If n_max > available eigenvalues all
                available ones are returned.
            k_points (int): At which k_point the eigenvalues are to be read.
            energy_units (str): Specify the units in which the energy is
                returned. Options: 'hartree', 'eV'

        Returns:
            list:
                List of eigenvalues (float).
        """

        dir_ = 'static'
        file = 'info'
        path = self.op._check_for_file(dir_, file)

        with open(path, 'r') as f:
            line = f.readline()

            if k_point is not None:
                # Periodic system
                while line and ('#k =' + f'{k_point}'.rjust(4)) not in line:
                    line = f.readline()

            else:
                # Non-periodic system
                while (line and
                       line != ' #st  Spin   Eigenvalue      Occupation\n'):
                    line = f.readline()

            if not line:
                print(f'{path} is not a valid \'info\' file.')
                raise

            eigenvalues = []
            if n_max is not None:
                line = f.readline()
                n = 1
                while (line and line != '\n' and '#k' not in line and
                       n <= n_max):

                    if n >= n_min:
                        eigenvalues.append(float(
                            re.search('-- *(.*?) ', line).groups()[0]))

                    line = f.readline()
                    n += 1

                if '#k' in line:
                    print('INFO: Not enough eigenvalues.')

            else:
                line = f.readline()
                n = 1
                while (line and line != '\n' and
                       '#k' not in line):

                    if n >= n_min:
                        eigenvalues.append(float(
                            re.search('-- *(.*?) ', line).groups()[0]))

                    line = f.readline()
                    n += 1

        eigenvalues = np.array(eigenvalues)
        if energy_units == 'eV':
            eigenvalues = units.energy(eigenvalues, from_='hartree', to='eV')

        return eigenvalues

    def get_cell(self):
        """Returns lattice vectors.

        Returns:
            nd.array:
                3x3 array where first index runs through the vectors
                and second through their coordinate.
        """

        dir_ = 'static'
        file = 'info'
        path = self.op._check_for_file(dir_, file)

        found = False
        with open(path) as f:
            for line in f:
                if line.startswith('  Lattice Vectors'):
                    lattice = np.zeros((3, 3))
                    for i in range(3):
                        lattice[i, :] = self._parse_vector(f.readline())
                        found = True
        if found:
            return lattice

        else:
            print('No reciprocal lattice found. Is this system periodic?')
            raise TypeError

    def get_reciprocal_cell(self):
        """Returns reciprocal lattice vectors.

        Returns:
            nd.array:
                3x3 array where first index runs through the vectors and second
                through their coordinate.
        """

        dir_ = 'static'
        file = 'info'
        path = self.op._check_for_file(dir_, file)

        found = False
        with open(path) as f:
            for line in f:
                if line.startswith('  Reciprocal-Lattice Vectors'):
                    lattice = np.zeros((3, 3))
                    for i in range(3):
                        lattice[i, :] = self._parse_vector(f.readline())
                        found = True
        if found:
            return lattice

        else:
            print('No reciprocal lattice found. Is this system periodic?')
            raise TypeError

    def get_convergence(self, plot=False):
        """Returns the convergence parameter energy and density.

        Args:
            plot (bool): If True a basic convergence plot for energy and
                density will be returned.

        Returns:
            nd.array:
                First column is the SCF cycle. Then energy, energy
                difference, density and relative density in this order.
            fig, ax, ax:
                Matplotlib handles for the figure, the left axis
                and the right axis in this order.
        """

        dir_ = 'static'
        file = 'convergence'
        path = self.op._check_for_file(dir_, file)
        data = np.loadtxt(path)[:, :5]

        if plot:
            fig, ax_left = plt.subplots()

            line_1 = ax_left.plot(data[:, 0], data[:, 1], marker='o',
                                  linestyle='', color='tab:red',
                                  label='Total Energy')
            line_2 = ax_left.plot(data[:, 0], data[:, 2], color='tab:red',
                                  label='Energy Difference', marker='x',
                                  linestyle='')

            ax_left.set_xlabel('SCF Cycle')
            ax_left.set_ylabel('Energy (Hartree)', color='tab:red')
            ax_left.tick_params(axis='y', labelcolor='tab:red')

            ax_right = ax_left.twinx()
            line_3 = ax_right.plot(data[:, 0], data[:, 3], marker='o',
                                   linestyle='', color='tab:blue',
                                   label='Total Density')
            line_4 = ax_right.plot(data[:, 0], data[:, 4], color='tab:blue',
                                   marker='x', linestyle='',
                                   label='Relative Density')

            ax_right.set_ylabel('Density', color='tab:blue')
            ax_right.tick_params(axis='y', labelcolor='tab:blue')

            lines = line_1 + line_2 + line_3 + line_4
            labels = [line.get_label() for line in lines]
            ax_left.legend(lines, labels, loc='center left')

            return data, (fig, ax_left, ax_right)

        else:
            return data

    def _parse_vector(self, line):
        vector = re.search(
            r' {3,4}([0-9\.\-]*) {3,4}([0-9\.\-]*) {3,4}([0-9\.\-]*)',
            line).groups((1, 2, 3))
        vector = [float(number) for number in vector]

        return vector



# -*- coding: utf-8 -*-
from pathlib import Path
from pytest import approx

import octopost


def test_get_eigenvalues():
    dir_path = Path(__file__).parent
    op = octopost.Octopost(dir_path / '../demo/data/graphene')

    eigenvalues = op.get_eigenvalues(n_max=None, k_point=1)

    assert(len(eigenvalues) == 5)

    eigenvalues = op.get_eigenvalues(n_max=1, k_point=3)
    assert(eigenvalues[0] == -0.867412)

    eigenvalues = op.get_eigenvalues(n_max=1, k_point=3, energy_units='eV')
    assert(eigenvalues[0] == approx(-23.60348296640494))

def test_get_fermi():
    dir_path = Path(__file__).parent
    op = octopost.Octopost(dir_path / '../demo/data/PTCDA')

    fermi = op.get_fermi()
    assert(fermi == approx(-0.240775))


# -*- coding: utf-8 -*-
"""Calculations related to the density of states.

This module contains the 'DOSModel' class, which handles all things
related to density of states calculations.
"""

# Python Imports
import re
import glob

# Third Party Imports
import numpy as np
import matplotlib.pyplot as plt

# Octopost Imports
import units


class DOSModel:
    """DOS model handles all things related to DOS.

    Should not be used directly by the user but called by the abstraction layer
    octopost class.

    Args:
        octopost (object): An instance of the octopost class.
    """

    def __init__(self, ocotpost):
        self.op = ocotpost

    def get_total_dos(self, plot=False, shift_by_fermi=True, plot_fermi=True,
                      energy_units='hartree'):
        """Extracts and possibly plots the total density of states.

        Reads the 'static/total-dos.dat' file, if available and returns the
        total density of states.

        Args:
            plot (bool): If True a basic plot will be returned.
            shift_by_fermi (bool): If True the energy will be shifted by the
                fermi energy (automatically extracted) for both the plot and
                the returned values.
            plot_fermi (bool): If True (and plot == True) a dashed line will
                be plotted as well where the fermi energy is.
            energy_units (str): Energy units to be plotted and returned.
                Options: 'hartree' (= Hartree), 'eV' (= Electron Volt)

        Returns:
            nd.array:
                2D array containing the enery in the units chosen in the first
                column and the density of states in the second.
            fig, ax:
                If plot == True the matplotlib handles for the plot will be
                returned as well.
        """

        file = 'total-dos.dat'
        dir_ = 'static'
        path = self.op._check_for_file(dir_, file)

        with open(path, 'r') as f:
            dos = np.loadtxt(f)

        fermi, dos, unit = self._dos_input_helper(
            dos, shift_by_fermi, plot_fermi, energy_units)

        if plot:
            fig, ax = self._dos_plot_helper(
                shift_by_fermi, plot_fermi, fermi, unit, axes=None)
            ax.set_title('Total Density of States')
            ax.plot(dos[:, 0], dos[:, 1], color='black', label='DOS')
            ax.legend()

            return dos, (fig, ax)

        else:
            return dos

    def get_band_dos(self, n_min=1, n_max=-1, plot=False,
                     shift_by_fermi=True, plot_fermi=True,
                     energy_units='hartree', axes=None):
        """Extracts and possibly plots the band projected density of states.

        Reads the band specific dos files in 'static', if available and
        returns the total densities of states.

        Args:
            n_min (int): The lowest band number to be extracted.
            n_max (int): The highest band number to be extracted. -1 means
                all bands larger n_min in the directory will be used.
            plot (bool): If True a basic plot will be returned.
            shift_by_fermi (bool): If True the energy will be shifted by the
                fermi energy (automatically extracted) for both the plot and
                the returned values.
            plot_fermi (bool): If True (and plot == True) a dashed line will
                be plotted as well where the fermi energy is.
            energy_units (str): Energy units to be plotted and returned.
                Options: 'hartree' (= Hartree), 'eV' (= Electron Volt)
            axes (fig, ax): Matplotlib handles of a already existing plot to
                overlay the information of this onto it.

        Returns:
            nd.array:
                (n_max - n_min + 1)D array containing the energy in the units
                chosen in the first column and the density of states projected
                onto state i in the ith column.
            fig, ax:
                If plot == True the matplotlib handles for the plot will be
                returned as well.
        """

        dir_ = 'static'
        if n_max == -1:
            aux = sorted(
                glob.glob(str(self.op.root_path / dir_ / 'dos-*.dat')))[-1]
            n_max = int(aux[-8:-4])

        for i in range(n_min, n_max + 1):
            file = 'dos-' + f'{i}'.zfill(4) + '.dat'
            path = self.op._check_for_file(dir_, file)
            with open(path, 'r') as f:
                data = np.loadtxt(f)

            if i == n_min:
                dos = data

            else:
                dos = np.column_stack((dos, data[:, 1]))

        fermi, dos, unit = self._dos_input_helper(
            dos, shift_by_fermi, plot_fermi, energy_units)

        if plot:
            fig, ax = self._dos_plot_helper(
                shift_by_fermi, plot_fermi, fermi, unit, axes)
            ax.set_title('Band Resolved Density of States')

            for i in range(n_max - n_min + 1):
                ax.plot(dos[:, 0], dos[:, i + 1], label=fr'Band {i + n_min}')

            ax.legend()

            return dos, (fig, ax)

        else:
            return dos

    def get_atomic_dos(self, atom, plot=False, shift_by_fermi=True,
                       plot_fermi=True, energy_units='hartree', axes=None,
                       combine_orbitals=True):
        """Extracts and possibly plots the atom projected density of states.

        Returns the density of states projected onto a specific atom.

        Args:
            atom (int): Number of the atom the DOS is to be returned.
            plot (bool): If True a basic plot will be returned.
            shift_by_fermi (bool): If True the energy will be shifted by the
                fermi energy (automatically extracted) for both the plot and
                the returned values.
            plot_fermi (bool): If True (and plot == True) a dashed line will
                be plotted as well where the fermi energy is.
            energy_units (str): Energy units to be plotted and returned.
                Options: 'hartree' (= Hartree), 'eV' (= Electron Volt)
            axes (fig, ax): Matplotlib handles of a already existing plot to
                overlay the information of this onto it.
            combine_orbitals (True): If True all the orbital projections will
                be summed up to a single density of states for this atom.

        Returns:
            nd.array:
                2D (if combine_orbitals) or (#orbitals of atom + 1)D array
                containing the energy in the units chosen in the first column
                and the density of states projected onto the orbitals or the
                entire atom in the remaining columns.
            fig, ax:
                If plot == True the matplotlib handles for the plot will be
                returned as well.
        """
        dir_ = 'static'
        paths = sorted(glob.glob(str(self.op.root_path /
                                     dir_ / ('pdos-at' +
                                             f'{atom}'.zfill(3) + '*.dat'))))

        orbitals = []
        for i in paths:
            orbitals.append(re.search(
                'pdos-at[0-9]{3}-(.*?).dat', i).groups()[0])

        for i, path in enumerate(paths):
            with open(path, 'r') as f:
                data = np.loadtxt(f)

            if i == 0:
                dos = data

            else:
                dos = np.column_stack((dos, data[:, 1]))

        fermi, dos, unit = self._dos_input_helper(
            dos, shift_by_fermi, plot_fermi, energy_units)

        if combine_orbitals:
            dos = np.column_stack((dos[:, 0], np.sum(dos[:, 1:], axis=1)))

        if plot:
            fig, ax = self._dos_plot_helper(
                shift_by_fermi, plot_fermi, fermi, unit, axes)
            ax.set_title(f'Atom {atom} Resolved Density of States')

            if combine_orbitals:
                ax.plot(dos[:, 0], dos[:, 1], label=fr'{atom}')

            else:
                for i in range(len(paths)):
                    ax.plot(dos[:, 0], dos[:, i + 1],
                            label=fr'Atom {atom} - Orbital {orbitals[i]}')

            ax.legend()

            return dos, (fig, ax)

        else:
            return dos

    def get_orbital_dos(self, orbital, plot=False, shift_by_fermi=True,
                        plot_fermi=True, energy_units='hartree', axes=None,
                        combine_atoms=True):
        """Extracts and possibly plots the orbital projected density of states.

        Returns and plots all the densities of states (possibly) combined
        projected onto a secific atom orbital.

        Args:
            orbital (str): Name of the atom orbital to be extracted in the
                form: 'element symbol' + 'energy level' + 'orbital type'.
                For example: Cu2p
            plot (bool): If True a basic plot will be returned.
            shift_by_fermi (bool): If True the energy will be shifted by the
                fermi energy (automatically extracted) for both the plot and
                the returned values.
            plot_fermi (bool): If True (and plot == True) a dashed line will
                be plotted as well where the fermi energy is.
            energy_units (str): Energy units to be plotted and returned.
                Options: 'hartree' (= Hartree), 'eV' (= Electron Volt)
            axes (fig, ax): Matplotlib handles of a already existing plot to
                overlay the information of this onto it.
            combine_atoms (bool): If true all atoms with such a orbital will
                be summed up to a single density of states.

        Returns:
            nd.array:
                2D (if combine_atoms) or (#orbitals of atom + 1)D array
                containing the energy in the units chosen in the first column
                and the density of states projected onto the orbitals or the
                entire atom in the remaining columns.
            fig, ax:
                If plot == True the matplotlib handles for the plot will be
                returned as well.
        """

        dir_ = 'static'
        paths = sorted(glob.glob(str(self.op.root_path / dir_ /
                                     ('pdos-at*' +
                                      f'{orbital}' + '.dat'))))

        atoms = []
        for i in paths:
            atoms.append(int(re.search(
                'pdos-at([0-9]{3})-.*?.dat', i).groups()[0]))

        for i, path in enumerate(paths):
            with open(path, 'r') as f:
                data = np.loadtxt(f)

            if i == 0:
                dos = data

            else:
                dos = np.column_stack((dos, data[:, 1]))

        fermi, dos, unit = self._dos_input_helper(
            dos, shift_by_fermi, plot_fermi, energy_units)

        if plot:
            fig, ax = self._dos_plot_helper(
                shift_by_fermi, plot_fermi, fermi, unit, axes)
            ax.set_title(f'Orbital {orbital} Resolved Density of States')

            if combine_atoms:
                dos[:, 1] = np.sum(dos[:, 1:], axis=1)
                ax.plot(dos[:, 0], dos[:, 1], label=fr'{orbital}')

            else:
                for i in range(len(paths)):
                    ax.plot(dos[:, 0], dos[:, i + 1],
                            label=fr'Atom {atoms[i]} - Orbital {orbital}')

            ax.legend()

            return dos, (fig, ax)

        else:
            return dos

    def _dos_input_helper(self, dos, shift_by_fermi, plot_fermi, energy_units):

        if shift_by_fermi or plot_fermi:
            fermi = self.op.get_fermi(energy_units)

        else:
            fermi = None

        if energy_units == 'eV':
            dos[:, 0] = units.energy(dos[:, 0], from_='hartree', to='eV')
            unit = 'eV'

        else:
            unit = 'Hartree'

        if shift_by_fermi:
            dos[:, 0] -= fermi
            fermi = 0

        return fermi, dos, unit

    def _dos_plot_helper(self, shift_by_fermi, plot_fermi, fermi, unit, axes):

        if axes is None:
            fig, ax = plt.subplots()

        else:
            fig, ax = axes

        if plot_fermi:
            ax.axvline(fermi, ls='--', color='lightgrey', label=r'$E_F$')

        if shift_by_fermi:
            ax.set_xlabel(fr'$E_B-E_F\;[\mathrm{{{unit}}}]$')

        else:
            ax.set_xlabel(fr'$E_B\;[\mathrm{{{unit}}}]$')
        ax.set_ylabel(r'$\mathrm{DOS}\;[\mathrm{arb. u.}]$')

        return fig, ax



# -*- coding: utf-8 -*-

import os
from pathlib import Path

from dos_model import DOSModel
from input_parser import Parser
from arpes_model import ARPESModel
from output_read_model import OutputReadModel
from bandstructure_model import BandstructureModel


class Octopost():

    def __init__(self, root=''):

        self.set_root(root)

    def set_root(self, root):
        self.root_path = self._validate_source_directory(root)
        self.parser = Parser(self)

    def get_bandstructure(self, *args, **kwargs):

        if not hasattr(self, 'bandstructure_model'):
            self.bandstructure_model = BandstructureModel(self)

        return self.bandstructure_model.get_bandstructure(*args, **kwargs)

    def get_fermi(self, *args, **kwargs):

        if not hasattr(self, 'output_read_model'):
            self.output_read_model = OutputReadModel(self)

        return self.output_read_model.get_fermi(*args, **kwargs)

    def get_eigenvalues(self, *args, **kwargs):

        if not hasattr(self, 'output_read_model'):
            self.output_read_model = OutputReadModel(self)

        return self.output_read_model.get_eigenvalues(*args, **kwargs)

    def get_convergence(self, *args, **kwargs):

        if not hasattr(self, 'output_read_model'):
            self.output_read_model = OutputReadModel(self)

        return self.output_read_model.get_convergence(*args, **kwargs)

    def get_cell(self, *args, **kwargs):

        if not hasattr(self, 'output_read_model'):
            self.output_read_model = OutputReadModel(self)

        return self.output_read_model.get_cell(*args, **kwargs)

    def get_reciprocal_cell(self, *args, **kwargs):

        if not hasattr(self, 'output_read_model'):
            self.output_read_model = OutputReadModel(self)

        return self.output_read_model.get_reciprocal_cell(*args, **kwargs)

    def get_total_dos(self, *args, **kwargs):

        if not hasattr(self, 'dos_model'):
            self.dos_model = DOSModel(self)

        return self.dos_model.get_total_dos(*args, **kwargs)

    def get_band_dos(self, *args, **kwargs):

        if not hasattr(self, 'dos_model'):
            self.dos_model = DOSModel(self)

        return self.dos_model.get_band_dos(*args, **kwargs)

    def get_atomic_dos(self, *args, **kwargs):

        if not hasattr(self, 'dos_model'):
            self.dos_model = DOSModel(self)

        return self.dos_model.get_atomic_dos(*args, **kwargs)

    def get_orbital_dos(self, *args, **kwargs):

        if not hasattr(self, 'dos_model'):
            self.dos_model = DOSModel(self)

        return self.dos_model.get_orbital_dos(*args, **kwargs)

    def get_gasphase_arpes(self, *args, **kwargs):

        if not hasattr(self, 'arpes_model'):
            self.arpes_model = ARPESModel(self)

        return self.arpes_model.get_gasphase_arpes(*args, **kwargs)

    def get_3D_arpes(self, *args, **kwargs):

        if not hasattr(self, 'arpes_model'):
            self.arpes_model = ARPESModel(self)

        return self.arpes_model.get_3D_arpes(*args, **kwargs)

    def get_state_resolved_arpes(self, *args, **kwargs):

        if not hasattr(self, 'arpes_model'):
            self.arpes_model = ARPESModel(self)

        return self.arpes_model.get_state_resolved_arpes(*args, **kwargs)

    def get_gasphase_pes(self, *args, **kwargs):

        if not hasattr(self, 'arpes_model'):
            self.arpes_model = ARPESModel(self)

        return self.arpes_model.get_gasphase_pes(*args, **kwargs)

    def get_3D_pes(self, *args, **kwargs):

        if not hasattr(self, 'arpes_model'):
            self.arpes_model = ARPESModel(self)

        return self.arpes_model.get_3D_pes(*args, **kwargs)

    def get_arpes_plot(self, *args, **kwargs):

        if not hasattr(self, 'arpes_model'):
            self.arpes_model = ARPESModel(self)

        return self.arpes_model.get_arpes_plot(*args, **kwargs)

    def _validate_source_directory(self, root):

        try:
            # TODO: Check validity
            return Path(root).resolve()

        except (RuntimeError, FileNotFoundError):
            print('Invalid source directory.')
            raise

    def _check_for_file(self, dir_, file):

        path = self.root_path / dir_ / file
        if os.path.isfile(path):
            return path

        else:
            print(f'No \'{file}\' file found in: {path}')
            raise



# -*- coding: utf-8 -*-
from pathlib import Path

import octopost


def test_runs():
    dir_path = Path(__file__).parent
    op = octopost.Octopost(dir_path / '../demo/data/graphene')

    op.get_bandstructure(plot=True)
    op.get_bandstructure(plot=False)



# -*- coding: utf-8 -*-
"""Calculations related to the ARPES maps.

This module contains the 'ARPESModel' class, which handles all things
related to ARPES.
"""

# Python Imports
import glob
import warnings

# Third Party Imports
import h5py
import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import RegularGridInterpolator

# Octopost Imports
import units
from library import (cartesian_to_spherical, E_to_k, k_to_E,
                     find_nearest_index, read_vtk_file)


class ARPESModel:
    """ARPES model handles all things related to ARPES.

    Should not be used directly by the user but called by the abstraction layer
    octopost class.

    Args:
        octopost (object): An instance of the octopost class.
    """

    def __init__(self, ocotpost):
        self.op = ocotpost

    def get_gasphase_arpes(self, file='PES_velocity_map.vtk',
                           E_num=10, E_range=[-10, 0], dE_b=0,
                           field='probe', omega=None, resolution=501,
                           out=False, out_file='output.hdf5'):
        """Calculates (kx, ky) ARPES maps at certain binding energies from a
        PES velocity file.

        Reads the PES velocity file specified with 'file' to calculate 'E_num'
        ARPES maps. The maps are calculated at binding energies specified in
        the semi open interval [E_range[0], E_range[1]). The energy of the
        incident light field can be specified via omega or read from the
        'parser.log' file by using the name 'field' the field was given.
        The resolution of each map is (resolution, resolution) and the data
        can be exported in a kMap.py compatible file 'out_file' using 'out'.

        Args:
            file (str): Relative path from the root directory of the
                calculation to the PES velocity map file (.vtk file).
            E_num (int): Number of energy cuts in the E_range interval.
            E_range (list): A list of length 2 specifying the semi-open
                interval [E_range[0], E_range[1]) of binding energies used for
                the energy cuts. All binding energies have to be E <= 0 and the
                units expected are 'eV'.
            dE_b (float): Integration over the open interval E_kin +- dE_b. In
                'eV'.
            field (str): Name the incident light field was given in the Octopus
                calculation. Is used to automatically find the photon energy of
                the incident light field. Can be overridden using 'omega'.
            omega (float): The energy of the incoming photon in 'eV'.
                Default is 'None', where Octopost tries to find the omega
                itself by reading the 'parser.log' file and looking up the
                field named 'field'.
            resolution (int): Resolution of the resulting maps. The maps have
                a equal resolution in kx and ky.
            out (bool): Whether or not to export the maps to a .hdf5 file. The
                file will contain the axis information and is compatible with
                kMap.
            out_file (str): If out is True the path to the output file
                relative to the root directory of the calculation.

        Returns:
            nd.array:
                A (E_num, resolution, resolution) shaped array. First
                index are the individual slices (even if E_num is 1). Second
                and third index are kx and ky respectively.
        Returns:
            nd.array:
                1D array containing the individual binding energies (not
                kinetic energy of the e) used for the slices in 'eV'.
            nd.array:
                1D array containing the kx (and therefore the identical
                ky) axis values in 1/Angstrom.
        """

        if omega is None:
            omega = self.op.parser.get_external_field(field)['omega']

        else:
            omega = units.energy(omega, from_='eV')

        E_range = units.energy(E_range, from_='eV')
        dE_b = units.energy(dE_b, from_='eV')
        E_axis = np.linspace(*E_range, E_num)
        # Maximum kinetic energy of photoelectron is energy of photon
        # plus the smallest (negative) binding energy
        E_kin_max = omega + max(E_range)
        slices = np.zeros((E_num, resolution, resolution))
        path = self.op._check_for_file('.', file)
        grid, intensity = read_vtk_file(path, array_name='PES_vel_map')
        # Grid from kx, ky, kz to v = |k|, theta, phi
        v, theta, phi = cartesian_to_spherical(
            grid[:, :, :, 0], grid[:, :, :, 1], grid[:, :, :, 2])
        v_axis = v[:, 0, 0]
        phi_axis = phi[0, :, 0]
        theta_axis = theta[0, 0, :]

        for i, E_b in enumerate(E_axis):
            E_kin = omega + E_b
            map_, k_axis = self._get_single_map(intensity, v_axis, theta_axis,
                                                phi_axis, E_kin, E_kin_max,
                                                dE_b, resolution)
            slices[i, :, :] = map_

        if out:
            self._write_sliced_data(out_file, slices, E_axis, k_axis)

        E_axis = units.energy(k_axis, from_='hartree')
        k_axis = units.inverse_length(k_axis, from_='bohr')

        return slices, E_axis, k_axis

    def get_3D_arpes(self, file='PES_velocity_map.vtk', E_num=10,
                     E_range=[-10, 0], dE_b=0, field='probe', omega=None,
                     out=False, out_file='output.hdf5'):
        """Calculates (kx, ky) ARPES maps at certain binding energies from a
        PES_ARPES file.

        Reads the PES ARPES file specified with 'file' to calculate 'E_num'
        ARPES maps. The maps are calculated at binding energies specified in
        the semi open interval [E_range[0], E_range[1]). The energy of the
        incident light field can be specified via omega or read from the
        'parser.log' file by using the name 'field' the field was given.
        The data can be exported in a kMap.py compatible file 'out_file'
        using 'out'.

        Args:
            file (str): Relative path from the root directory of the
                calculation to the PES velocity map file (.vtk file).
            E_num (int): Number of energy cuts in the E_range interval.
            E_range (list): A list of length 2 specifying the semi-open
                interval [E_range[0], E_range[1]) of binding energies used for
                the energy cuts. All binding energies have to be E <= 0 and the
                units expected are 'eV'.
            dE_b (float): Integration over the open interval E_kin +- dE_b. In
                'eV'.
            field (str): Name the incident light field was given in the Octopus
                calculation. Is used to automatically find the photon energy of
                the incident light field. Can be overridden using 'omega'.
            omega (float): The energy of the incoming photon in 'eV'.
                Default is 'None', where Octopost tries to find the omega
                itself by reading the 'parser.log' file and looking up the
                field named 'field'.
            out (bool): Whether or not to export the maps to a .hdf5 file. The
                file will contain the axis information and is compatible with
                kMap.
            out_file (str): If out is True the path to the output file
                relative to the root directory of the calculation.

        Returns:
            nd.array:
                A (E_num, resolution, resolution) shaped array. First
                index are the individual slices (even if E_num is 1). Second
                and third index are kx and ky respectively.
            nd.array:
                1D array containing the individual binding energies (not
                kinetic energy of the e) used for the slices in 'eV'.
            nd.array:
                1D array containing the kx (and therefore the identical
                ky) axis values in 1/Angstrom.
        """

        if omega is None:
            omega = self.op.parser.get_external_field(field)['omega']

        else:
            omega = units.energy(omega, from_='eV')

        E_range = units.energy(E_range, from_='eV')
        dE_b = units.energy(dE_b, from_='eV')
        Ebs = np.linspace(*E_range, E_num)

        path = self.op._check_for_file('.', file)
        grid, intensity = read_vtk_file(path, array_name='PES_vel_map')
        kx_axis = grid[:, 0, 0][:, 0]
        ky_axis = grid[0, :, 0][:, 1]
        E_axis = grid[0, 0, :][:, 2]

        slices = np.zeros((E_num, len(kx_axis), len(ky_axis)))
        for i, E_b in enumerate(Ebs):
            E_kin = omega + E_b
            map_ = self._cut_energy_3D(intensity, E_axis, E_kin, dE_b)
            slices[i, :, :] = map_

        if out:
            self._write_sliced_data(out_file, slices, E_axis, kx_axis)

        E_axis = units.energy(E_axis, from_='hartree')
        kx_axis = units.inverse_length(kx_axis, from_='bohr')
        ky_axis = units.inverse_length(ky_axis, from_='bohr')

        return slices, Ebs, kx_axis

    def get_state_resolved_arpes(self, states=-1, field='probe', omega=None,
                                 dE_b=0.02, resolution=501):
        """Calculate a state projected (kx, ky) ARPES map for each state.

        Using the state projected photoemission intensity files outputed by
        Octopus (if requested), for each state in 'states' the state projected
        ARPES map is calculated. The binding energy will be taken from the
        eigenvalues of the indiviudal states.

        Args:
            states (list): List of integers specifying the states to be used.
                For each entry a corresponding file has to be present. If -1
                all files found will be used.
            dE_b (float): See 'get_gasphase_arpes'.
            field (str): See 'get_gasphase_arpes'.
            omega (float): See 'get_gasphase_arpes'.
            resolution (int): See 'get_gasphase_arpes'.

        Returns:
            nd.array:
                A (E_num, resolution, resolution) shaped array. First
                index are the individual maps (even if len(states) is 1).
                Second and third index are kx and ky respectively.
            nd.array:
                1D array containing the individual binding energies (not
                kinetic energy of the e) used for the states in 'eV'.
            nd.array:
                1D array containing the kx (and therefore the identical
                ky) axis values in 1/Angstrom.
        """

        if states == -1:
            aux = sorted(
                glob.glob(str(self.op.root_path /
                              'PES_velocity_map-st*.vtk')))
            states = [int(i[-8:-4]) for i in aux]

        E_bs = self.op.get_eigenvalues(n_max=states[-1], n_min=states[0],
                                       energy_units='eV')

        slices = np.zeros((len(states), resolution, resolution))
        for i, (state, E_b) in enumerate(zip(states, E_bs)):
            file = f'PES_velocity_map-st{state:04}.vtk'
            slice_, E_axis, k_axis = self.get_gasphase_arpes(
                file=file, E_num=1, E_range=[E_b, E_b], dE_b=dE_b,
                field=field, omega=omega, resolution=resolution)
            slices[i, :, :] = slice_

        return slices, np.array(E_bs), k_axis

    def get_gasphase_pes(self, file, integrate=('theta', 'phi'), plot=False):
        """Returns photoemission intensity.

        Reads the photoemission intensity from 'file' and, if requested,
        integrates over 0,1, 2 or all dimensions. Option to plot is only
        available for integration over 2 dimensions.

        Args:
            file (str): Relative path from the root directory of the
                calculation to the PES velocity map file (.vtk file).
            integrate (list): List containing any combination of 'theta',
                'phi' and 'energy' including empty and all of them. The
                dimensions specified will be integrated over.
            plot (bool): If len(integrate) != 2 this option will be ignored.
                Otherwise a basic line plot over the remaining dimension will
                be returned.

        Returns:
            nd.array or float:
                The intensity integrated over the dimensions
                specified in arbitrary units. If integrated over all
                dimensions this will be a float, otherwise a nd.array with
                3 - len(integrate) dimensions. The first dim will be 'energy'
                or, if integrated over, 'phi'. The last dim will be 'theta' or,
                if integrated over, 'phi'.
            tuple or nd.array:
                Remaining axes in the correct order. If only
                one remains this will be a nd.array, if none remains no return
                value and otherwise a tuple of axes.
            fig, ax:
                If plot == True and integration over two dimensions, the
                handles for the matplotlib plot.
        """

        path = self.op._check_for_file('.', file)
        grid, intensity = read_vtk_file(path, array_name='PES_vel_map')
        v, theta, phi = cartesian_to_spherical(
            grid[:, :, :, 0], grid[:, :, :, 1], grid[:, :, :, 2])
        E_axis = units.energy(k_to_E(v[:, 0, 0]), from_='hartree')
        phi_axis = phi[0, :, 0]
        theta_axis = theta[0, 0, :]

        if 'theta' in integrate:
            intensity = np.sum(intensity, axis=2)
            theta_axis = False

        if 'phi' in integrate:
            intensity = np.sum(intensity, axis=1)
            phi_axis = False

        if 'energy' in integrate:
            intensity = np.sum(intensity, axis=0)
            E_axis = False

        if np.isscalar(intensity):
            return intensity

        elif intensity.ndim == 3:
            return intensity, (E_axis, phi_axis, theta_axis)

        elif intensity.ndim == 2:

            if theta_axis is False:
                return intensity, (E_axis, phi_axis)

            elif phi_axis is False:
                return intensity, (E_axis, theta_axis)

            else:
                return intensity, (phi_axis, theta_axis)

        else:
            if plot:
                fig, ax = plt.subplots()
                ax.set_ylabel('$I$ (a.u.)', fontsize=14)

                if theta_axis is not False:
                    ax.plot(theta_axis, intensity)
                    ax.set_xlabel(r'$\theta$', fontsize=14)

                    return intensity, theta_axis, (fig, ax)

                if phi_axis is not False:
                    ax.plot(phi_axis, intensity)
                    ax.set_xlabel(r'$\phi$ (eV)', fontsize=14)

                    return intensity, phi_axis, (fig, ax)

                else:
                    ax.plot(E_axis, intensity)
                    ax.set_xlabel(r'$E_{kin}$ (eV)', fontsize=14)

                    return intensity, E_axis, (fig, ax)

            else:
                return intensity, E_axis

    def get_3D_pes(self, file, plot=False):
        """Returns photoemission intensity over the energy.

        Args:
            file (str): Relative path from the root directory of the
                calculation to the PES velocity map file (.vtk file).
            plot (bool): If len(integrate) != 2 this option will be ignored.
                Otherwise a basic line plot over the remaining dimension will
                be returned.

        Returns:
            nd.array:
                The intensity integrated over the kx and ky axes.
            nd.array:
                The energy axis in eV.
                fig, ax: If plot == True and integration over two dimensions,
                the handles for the matplotlib plot.
        """

        path = self.op._check_for_file('.', file)
        grid, intensity = read_vtk_file(path, array_name='PES_vel_map')
        E_axis = units.energy(grid[0, 0, :][:, 2], from_='hartree')

        intensity = np.sum(intensity, axis=(0, 1))

        if plot:
            fig, ax = plt.subplots()
            ax.set_xlabel(r'$E_{kin}$ (eV)', fontsize=14)
            ax.set_ylabel('$I$ (a.u.)', fontsize=14)
            ax.plot(E_axis, intensity)

            return intensity, E_axis, (fig, ax)

        else:
            return intensity, E_axis

    def get_arpes_plot(self, map_, k_axis):
        """Convenience method to plot a ARPES map.

        Uses the output of 'get_gasphase_arpes' to quickly plot an ARPES map.

        Args:
            map_ (nd.array): 2D array containing the ARPES map.
            k_axis (nd.array): 1D array containing the kx and ky axis in 1/A.

        Returns:
            fig, ax:
                Matplotlib handles for the plot.
        """

        fig, ax = plt.subplots()
        ax.imshow(map_, extent=(k_axis[0], k_axis[-1], k_axis[0], k_axis[-1]))
        ax.set_xlabel('$k_x$ (1/A)', fontsize=14)
        ax.set_ylabel('$k_y$ (1/A)', fontsize=14)

        return fig, ax

    def _get_single_map(self, intensity, v_axis, theta_axis, phi_axis, E_kin,
                        E_kin_max, dE_b, resolution):
        """Private helper method to calculate APRES from for single binding
        energy.

        Should not be used directly. Please refer to the 'get_gasphase_arpes'
        function instead.
        """

        intensity_cut = self._cut_energy_gasphase(intensity, v_axis, E_kin,
                                                  dE_b)
        phi_axis_connect, intensity_cut = self._connect_phi(
            phi_axis, theta_axis, intensity_cut)
        map_, k_axis = self._interp_spherical_to_cartesian(
            intensity_cut, theta_axis, phi_axis_connect,
            E_kin_max, resolution)

        return map_, k_axis

    def _cut_energy_gasphase(self, intensity, v_axis, E_kin, dE_b=0):
        """Private helper method to cut a single energy slice from data.

        Cuts a hemisphere at E_kin from the intensity data. If dE_b is not 0,
        then there will be a integration of a the radius E_kin +- dE_b.
        Should not be used directly. Please refer to the 'get_gasphase_arpes'
        function instead.
        """

        k_at_cut = E_to_k(E_kin)
        if dE_b != 0:
            dk = E_to_k(dE_b)
            mask = np.where((v_axis < (k_at_cut + dk)) &
                            (v_axis > (k_at_cut - dk)))[0]

        if dE_b == 0 or len(mask) == 0:
            if dE_b != 0:
                # => len(mask) == 0
                print('WARNING: No energy slice inside (E-dE, E+dE).')

            idx = find_nearest_index(v_axis, k_at_cut)
            return intensity[idx, :, :]

        return np.sum(intensity[mask, :, :], axis=0)

    def _cut_energy_3D(self, intensity, E_axis, E_kin, dE_b=0):
        """Private helper method to cut a single energy slice from data.

        Cuts at a certain E_kin from the intensity data. If dE_b is not 0,
        then there will be a integration of a the radius E_kin +- dE_b.
        Should not be used directly. Please refer to the 'get_3D_arpes'
        function instead.
        """

        if dE_b != 0:
            mask = np.where((E_axis < (E_kin + dE_b)) &
                            (E_axis > (E_kin - dE_b)))[0]

        if dE_b == 0 or len(mask) == 0:
            if dE_b != 0:
                # => len(mask) == 0
                print('WARNING: No energy slice inside (E-dE, E+dE).')

            idx = find_nearest_index(E_axis, E_kin)
            return intensity[:, :, idx]

        return np.sum(intensity[:, :, mask], axis=2)

    def _write_sliced_data(self, file, slices, E_axis, k_axis):
        """Private helper method to write a kMap.py combatible .hdf5 file.

        Should not be used directly. Please refer to the 'get_gasphase_arpes'
        function instead.
        """

        with h5py.File(self.op.root_path / file, 'w') as f:
            f.create_dataset('name', data='Test Export')
            f.create_dataset('axis_1_label', data='Binding Energy')
            f.create_dataset('axis_2_label', data='k_x')
            f.create_dataset('axis_3_label', data='k_y')
            f.create_dataset('axis_1_units', data='eV')
            f.create_dataset('axis_2_units', data='1/A')
            f.create_dataset('axis_3_units', data='1/A')

            axis_1_range = units.energy([np.nanmin(E_axis), np.nanmax(E_axis)])
            axis_2_range = units.inverse_length(
                [np.nanmin(k_axis), np.nanmax(k_axis)])
            f.create_dataset('axis_1_range', data=axis_1_range)
            f.create_dataset('axis_2_range', data=axis_2_range)
            f.create_dataset('axis_3_range', data=axis_2_range)
            f.create_dataset(
                'data', data=slices, dtype='f8', compression='gzip',
                compression_opts=9)

        return

    def _connect_phi(self, phi_axis, theta_axis, intensity):
        """Private helper method to complete the phi coordinate.

        The phi coordinate runs in the interval [0, 2pi). The points at 2pi
        has no further information (is identical to 0) but for plotting the
        phi coordinate should run in a closed interval.
        Should not be used directly. Please refer to the 'get_gasphase_arpes'
        function instead.
        """

        phi_axis = np.append(phi_axis, 2 * np.pi)
        intensity = np.vstack((intensity, intensity[0, :]))

        return phi_axis, intensity

    def _interp_spherical_to_cartesian(self, intensity, theta_axis,
                                       phi_axis, E_max, resolution):
        """Private helper method to interpolate a hemispherical energy cut
        down to a regular x,y grid.

        Should not be used directly. Please refer to the 'get_gasphase_arpes'
        function instead.
        """

        interp = RegularGridInterpolator(
            (phi_axis - np.pi, theta_axis), intensity,
            bounds_error=False, fill_value=np.nan)
        k_max = E_to_k(E_max)
        # x and y axis for the new kx, ky grid
        k_axis = np.linspace(-k_max, k_max, resolution)
        KX, KY = np.meshgrid(k_axis, k_axis)
        # k_max is equivalent to the radius of the energy cut
        # Spherical coordinates for the new grid
        phi_new = np.arctan2(KY, KX)
        k = np.sqrt(KX**2 + KY**2)

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            theta_new = np.arcsin(k / k_max)

        points = np.reshape(np.stack((phi_new, theta_new)), (2, -1)).T
        map_ = np.reshape(interp(points), (len(k_axis), len(k_axis)))

        return np.rot90(map_, 2), k_axis



from src.__init__ import __version__
from setuptools import setup, find_packages

DESCRIPTION = 'Octopost contains a pool of python3 tools for post-processing data obtained with the TDDFT code OCTOPUS[1].'

with open('README.rst', 'r') as fh:
    LONG_DESCRIPTION = fh.read()

setup(
    name='Octopost',
    version=__version__,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/x-rst',
    url='https://gitlab.com/ckern/octopost',
    author='Christian Kern',
    author_email='christian.kern@uni-graz.at',
    package_dir={'': 'src'},
    packages=[''],
    python_requires='>=3.8',
    install_requires=['numpy>=1.20', 'scipy>=1.5', 'vtk>=9.0',
                      'matplotlib>=3.3', 'h5py>=3.0'],
    extras_require={
        'test': ['pytest==6.2.5', 'flake8==3.9.2'],
        'dev': ['sphinx==4.1.2', 'check-manifest==0.46',
                       'twine==3.4.2', 'sphinx-rtd-theme==0.5.2']},
    classifiers=['Development Status :: 4 - Beta',
                 'Topic :: Documentation :: Sphinx',
                 'Natural Language :: English',
                 'Operating System :: POSIX :: Linux',
                 'Programming Language :: Python',
                 'Programming Language :: Python :: 3',
                 'Programming Language :: Python :: 3.8',
                 'Topic :: Communications :: Email',
                 'Topic :: Scientific/Engineering :: Physics',
                 'Topic :: Software Development :: Version Control :: Git',
                 'Topic :: Text Processing :: Markup :: reStructuredText',
                 'License :: OSI Approved :: MIT License'
                 ]
)



# -*- coding: utf-8 -*-
"""Parser for the input of the Octopus calculation.

A small text parser specifically for the 'parser.log' file of a Octopus
calculation containing the inputs done by the user. Used for automatic
extraction of certain values necessary for the data post processing.
"""

# Python Imports
import re


class Parser():
    """Textparser class.

    Args:
        octopost (object): An instance of the octopost class.

    Attributes:
        input (dict): A dictionary containing all parsed variables from the
            parser.log file. Public to directly access not yet abstracted
            variables.
    """

    def __init__(self, ocotpost):

        self.op = ocotpost
        self.input = {}

        self._parse()

    def get_external_field(self, field):
        """Returns the external filed specified in the Octopus calculation.

        Args:
            field (str): Each field in a Octopus calculation is given a name.
                This name is used to find the field data.

        Returns:
            dict:
                All available data to this field in form of a dictionary. The
                available keys depend on the type of field. For more
                information please see the Octopus documentation:
                https://octopus-code.org/doc/11.1/html/vars.php?page=alpha
        """

        field_key = None
        field = f'"{field}"'
        for key, _ in self.input.items():
            if key.startswith('TDExternalFields') and self.input[key] == field:
                field_key = key
                name_index = re.search(
                    r'TDExternalFields\[[0-9]+\]\[([0-9]+)\]',
                    field_key).group(1)
                type_ = 'scalar' if name_index == 2 else 'vector'

        if field_key is None:
            print('ERROR: Field not found in parser.log file.')
            raise ValueError

        else:
            field_number = re.search(
                r'TDExternalFields\[([0-9]+)\]', field_key).group(1)

        if type_ == 'scalar':
            keys = ('scalar_potential', 'spatial_expression',
                    'omega', 'envelope_function_name', 'phase')
        else:
            keys = ('type', 'nx', 'ny', 'nz', 'omega',
                    'envelope_function_name', 'phase')

        field_dict = {}
        for i, key in enumerate(keys):
            try:
                field_dict.update({key: self.input[
                    f'TDExternalFields[{field_number}][{i}]']})

            except KeyError:
                if key == 'phase':
                    # Phase is an optional parameter
                    pass
                else:
                    print(f'The key "{key}" was has not been found for the '
                          f'external field {field}')
                    raise KeyError

        return field_dict

    def _parse(self):

        file = 'parser.log'
        dir_ = 'exec'
        path = self.op._check_for_file(dir_, file)

        with open(path) as f:
            for line in f:
                if line == '' or line[0] == '#':
                    pass

                else:
                    aux = self._parse_line(line)

                    if aux:
                        self.input.update(aux)

    def _parse_line(self, line):

        if('=' in line):
            groups = re.search('(.*) = (.+?)[\t+|\n#]', line).groups()
            key = groups[0]

            try:
                value = int(groups[1])

            except ValueError:
                try:
                    value = float(groups[1])

                except ValueError:
                    value = groups[1]

            return {key: value}

        else:
            return False
